/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */

import 'dart:io';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:glob/glob.dart';
import 'package:source_gen/source_gen.dart';
import 'package:welltested/annotation.dart';
import 'package:welltested/src/executors/ast_elm_executor.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/extensions/annotation_extension.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/repositories/path_repository.dart';
import 'package:welltested/src/utilities/utilities.dart';

import '../src/visitors/class_visitor.dart';

class SaveBuilder extends Builder {
  late final DartServerRepository _dartServerRepository;
  late final PathRepository _pathRepository;
  late final FixerExecutor _fixerExecutor;
  late final GenerationRepository _generationRepository;
  late final AstElementExecutor _astElementExecutor;
  late final UnitTestExecutor _unitTestExecutor;
  late final BuildStep _buildStep;

  static final _codeFiles = Glob('lib/**.dart');
  static final _testFiles = Glob('test/**.welltested_test.dart');

  static AssetId _allFileOutput(BuildStep buildStep) {
    return AssetId(
      buildStep.inputId.package,
      'lib${Platform.pathSeparator}save_schema.txt',
    );
  }

  @override
  Map<String, List<String>> get buildExtensions {
    return const {
      r'$package$': ['lib/save_schema.txt'],
    };
  }

  @override
  Future<void> build(BuildStep buildStep) async {
    _buildStep = buildStep;

    Map<AssetId, String> testFiles = {};

    //Repositories
    _dartServerRepository = DartServerRepository();
    await _dartServerRepository.initialize();
    _pathRepository = PathRepository();
    _generationRepository = GenerationRepository();

    ///Executors
    _fixerExecutor = FixerExecutor(_dartServerRepository
        .server); //TODO: Create methods and don't pass server directly
    _astElementExecutor = AstElementExecutor();
    _unitTestExecutor = UnitTestExecutor(_generationRepository);

    await for (final input in buildStep.findAssets(_testFiles)) {
      final testContent = await buildStep.readAsString(input);
      testFiles[input] = testContent;
    }
    final List<Future<Map<String, dynamic>>> saveFutures = [];

    await for (final input in buildStep.findAssets(_codeFiles)) {
      try {
        final inputLibrary =
            await buildStep.resolver.libraryFor(input, allowSyntaxErrors: true);

        for (final topLevelElement in inputLibrary.topLevelElements) {
          final checker = TypeChecker.fromRuntime(Welltested);
          final annotationElement = checker.firstAnnotationOf(
            topLevelElement,
            throwOnUnresolved: true,
          );

          if (annotationElement != null) {
            if (!(topLevelElement is ClassElement)) return;

            try {
              saveFutures.addAll(await saveTests(
                  topLevelElement, input, ConstantReader(annotationElement)));
            } catch (e) {
              log.severe(
                  "Error saving tests for ${topLevelElement.name} in ${input.path}");
            }
          }
        }
      } catch (e) {}
    }
    await Future.wait(saveFutures);
    _dartServerRepository.server.dispose();

    ///TODO: can save some metadata here
    return buildStep.writeAsString(_allFileOutput(buildStep), '//Placeholder');
  }

  Future<List<Future<Map<String, dynamic>>>> saveTests(
      ClassElement element, AssetId assetId, ConstantReader annotation) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final classAstNode = await getResolvedAstNodeForElement(element, _buildStep)
        as ClassDeclaration; //Orignally:  (await _buildStep.resolver.astNodeFor(element)) as ClassDeclaration;
    final fileAstNode = classAstNode.parent;
    final fileImport =
        'package:${assetId.package}/${assetId.path.replaceAll('lib/', '')}';

    final classVisitor = ClassVisitor();

    element.visitChildren(classVisitor);

    final Map<String, MethodDeclaration> methodAstNodes =
        await _astElementExecutor.methodNodesFromClassVisitor(
            _buildStep, classVisitor);
    final List<Future<Map<String, dynamic>>> saveFutures = [];
    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      final testRequestData = await _unitTestExecutor.prepareRequest(
          method: method,
          fileAstNode: fileAstNode,
          classAstNode: classAstNode,
          fileImport: fileImport,
          className: classVisitor.className,
          methodAstNodes: methodAstNodes,
          buildStep: _buildStep,
          excludedMethods: excludedMethods);
      if (testRequestData == null) continue;

      final file = await _pathRepository.correspondingPermanentTestFile(
          method.key, assetId);

      if (file == null) continue;

      var hasErrors =
          await _fixerExecutor.hasCompilationErrors(file.absolute.path);

      if (hasErrors) {
        log.info(
            "⚠️ File ${file.path.split(Platform.pathSeparator).last} has errors and will not be updated");
        continue;
      }
      saveFutures.add(_unitTestExecutor.updateTest(
          testRequestData, await file.readAsString()));
    }
    return saveFutures;
  }
}
