/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'dart:async';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';
import 'package:welltested/annotation.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/executors/writer_executor.dart';
import 'package:welltested/src/extensions/annotation_extension.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/repositories/path_repository.dart';
import 'package:welltested/src/repositories/writer_repository.dart';
import 'package:welltested/src/visitors/class_visitor.dart';

import '../src/executors/ast_elm_executor.dart';

class WelltestedGenerator extends GeneratorForAnnotation<Welltested> {
  bool initialized = false;

  @override
  Future<FutureOr<String>> generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) async {
    if (element is! ClassElement) {
      final String friendlyName = element.displayName;
      throw InvalidGenerationSourceError(
        'Generator cannot target `$friendlyName`.',
        todo: 'Remove the [Welltested] annotation from `$friendlyName`.',
      );
    }
    //Repositories

    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();
    final pathRepository = PathRepository();
    final generationRepository = GenerationRepository();
    final writerRepository = WriterRepository(buildStep);

    ///Executors
    final fixerExecutor = FixerExecutor(dartServerRepository
        .server); //TODO: Create methods and don't pass server directly
    final astElementExecutor = AstElementExecutor();
    final unitTestExecutor = UnitTestExecutor(generationRepository);

    final writerExecutor = WriterExecutor(
        pathRepository, writerRepository, dartServerRepository, fixerExecutor);
    final List<String> excludedMethods = annotation.excludeMethods;

    final classAstNode =
        (await buildStep.resolver.astNodeFor(element)) as ClassDeclaration;
    final fileAstNode = classAstNode.parent;
    final fileImport =
        'package:${buildStep.inputId.package}/${buildStep.inputId.path.replaceAll('lib/', '')}'; //TODO: Extract to extension
    final classVisitor = ClassVisitor();
    element.visitChildren(classVisitor);
    final methodAstNodes = await astElementExecutor.methodNodesFromClassVisitor(
        buildStep, classVisitor); //TODO: probably an extension is better fit
    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      try {
        final method = methodAstNodes.entries.elementAt(i);
        final testRequestData = await unitTestExecutor.prepareRequest(
            method: method,
            fileAstNode: fileAstNode,
            classAstNode: classAstNode,
            fileImport: fileImport,
            className: classVisitor.className,
            methodAstNodes: methodAstNodes,
            buildStep: buildStep,
            excludedMethods: excludedMethods);
        if (testRequestData == null) continue;
        log.info(" Generating Test for ${testRequestData.functionName}");
        final file = await pathRepository.correspondingPermanentTestFile(
            method.key, buildStep.inputId);
        if (file == null) {
          testRequestData.forceGenerate = true;
        }
        final unitTests =
            await unitTestExecutor.buildUnitTests(testRequestData);

        final overrideResult = unitTests['fixed_completion'];
        if (overrideResult != null) {
          if (file != null && (file.readAsString() == overrideResult)) {
            continue;
          }
          await writerExecutor.outputGeneration(
              i, method.key, overrideResult, buildStep.inputId);
          continue;
        }
        if (unitTests['model_completion'] == null) {
          log.warning(
              "⛔️ Sorry, we can't process ${testRequestData.className} at the moment. Please try again Later.");
          continue;
        }
        await writerExecutor.processGeneration(
            buildStep.inputId, unitTests['model_completion'], method.key, i);
      } catch (e) {
        log.severe(
            "Error generating unit tests for method: ${methodAstNodes.entries.elementAt(i).key}");
      }
    }

    dartServerRepository.server.dispose();
    return '';
  }

  @override
  FutureOr<String> generate(LibraryReader library, BuildStep buildStep) async {
    //TODO: Generate is called for every file on the same instance of builder.
    //Hence variables can't be initialized twice.
    return super.generate(library, buildStep);
  }
}
