/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'dart:io';

import 'package:analysis_server_lib/analysis_server_lib.dart';
import 'package:analysis_server_lib/analysis_server_lib.dart' as serverlib;
import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/file_system/memory_file_system.dart';
import 'package:build/build.dart';
import 'package:welltested/src/extensions/analysis_errors_extenstion.dart';
import 'package:welltested/src/extensions/replace_list_extension.dart';
import 'package:welltested/src/visitors/unit_test_visitor.dart';

class FixerExecutor {
  final serverlib.AnalysisServer server;
  FixerExecutor(this.server);

  Future<void> setRoots(List<String> includedFiles) async {
    await server.analysis.setAnalysisRoots(includedFiles, []);
  }

  Future<bool> hasCompilationErrors(String filePath) async {
    await server.analysis.setAnalysisRoots([filePath],
        []); //TODO: Shift to Dart Server repository add files to root.
    final analysis = await server.analysis.getErrors(filePath);
    bool hasErrors = false;
    for (var error in analysis.errors) {
      if (error.severity == 'ERROR') {
        hasErrors = true;
        break;
      }
    }
    return hasErrors;
  }

  Future<String> findAndFixIssues(
      String filePath, String functionIdentifer) async {
    File file = File(filePath);
    String fileContent = await file.readAsString();
    var analysis = await server.analysis.getErrors(filePath);
    String content = '';
    content = await _fixImportIssues(
        analysis.errors.fixableUndefinedErrors, filePath, fileContent, file);
    content = addMocksFile(content, functionIdentifer);

    return content;
  }

  String addMocksFile(String content, String functionidentifier) {
    if (!content.contains("@GenerateMocks([")) {
      return content;
    }
    return content.replaceRange(
        0, 0, "import '$functionidentifier.welltested_test.mocks.dart';");
  }

  Future<String> _fixImportIssues(List<AnalysisError> undefinedErrors,
      String filePath, String fileContent, File file) async {
    final List<SourceEdit> replacements = [];
    for (final error in undefinedErrors) {
      final FixesResult fix = await server.edit.getFixes(
        filePath,
        error.location.offset,
      );

      if (fix.fixes.isNotEmpty) {
        Set<SourceEdit> _fixes = Set();
        for (var element in fix.fixes) {
          for (var e in element.fixes) {
            if (e.id?.contains("dart.fix.import") ?? false) {
              _fixes.add(e.edits.first.edits.first);
            }
          }
        }
        replacements.addAll(_fixes);
      }
    }
    final updatedContent = _runSourceEditSequence(fileContent, replacements);
    await file.writeAsString(updatedContent, flush: true);
    return updatedContent;
  }

  ///Run analysis on code that hasn't been written to a file yet
  Future<ResolvedUnitResult> runAnalysisOnGenerated(
      {required LibraryElement element,
      required String path,
      required BuildStep buildStep}) async {
    String generated = await File(path).readAsString();
    final MemoryResourceProvider resource =
        element.session.resourceProvider as MemoryResourceProvider;
    resource.newFile(path, generated);
    final ResolvedLibraryResult fileAnalysis =
        await element.session.getResolvedLibrary(path) as ResolvedLibraryResult;
    final astNode =
        await buildStep.resolver.astNodeFor(fileAnalysis.element.units.first);
    astNode?.visitChildren(UnitTestVisitor());
    return fileAnalysis.units.first;
  }

  Future<String> fixRepeatingIssues(
      {required LibraryElement element,
      required String path,
      required BuildStep buildStep}) async {
    String generated = await File(path).readAsString();
    final MemoryResourceProvider resource =
        element.session.resourceProvider as MemoryResourceProvider;
    resource.newFile(path, generated);
    final ResolvedLibraryResult fileAnalysis =
        await element.session.getResolvedLibrary(path) as ResolvedLibraryResult;
    final astNode =
        await buildStep.resolver.astNodeFor(fileAnalysis.element.units.first);
    if (astNode == null) return 'Couldn\'t fix repeating issues';
    final visitor = UnitTestVisitor();
    astNode.visitChildren(visitor);
    final updatedContent = _runReplacements(generated, visitor.replacements);
    await File(path).writeAsString(updatedContent, flush: true);
    return updatedContent;
  }

  String _runSourceEditSequence(String content, List<SourceEdit> sourceEdits) {
    sourceEdits.sort((a, b) => a.offset.compareTo(b.offset));
    sourceEdits = sourceEdits.filterDuplicationReplacements;
    String updatedContent = content;
    int shift = 0;
    for (final edit in sourceEdits) {
      updatedContent = updatedContent.replaceRange(edit.offset + shift,
          edit.offset + edit.length + shift, edit.replacement);
      shift = shift + edit.replacement.length - edit.length;
    }
    return updatedContent;
  }

  ///If the selections overlap, this function will fail. TODO Update to handle that.
  String _runReplacements(String content, List<ReplaceModel> replaceModels) {
    String updatedContent = content;
    replaceModels = replaceModels.extractUniqueAndUnoverlapped;
    replaceModels.sort((a, b) => a.offset.compareTo(b.offset));
    int shift = 0;

    for (final model in replaceModels) {
      updatedContent = updatedContent.substring(0, model.offset + shift) +
          model.replacement +
          updatedContent.substring(model.offset + model.length + shift);
      shift = shift + model.replacement.length - model.length;
    }

    return updatedContent;
  }
}
